//|------------------------------------------------------------------+
//|																													 SOM_ex5 |
//|												 Copyright 2011, MetaQuotes Software Corp. |
//|															 								 http://www.mql5.com |
//|------------------------------------------------------------------+
#property copyright "Copyright 2011, MetaQuotes Software Corp."
#property link		"http://www.mql5.com"
#property version	"1.00"

#include "cIntBMP.mqh" //   cIntBMP
#include "SOMNode.mqh"

enum ColorSchemes
{
	Black_White=0,		// -
	Blue_Green_Red=1,	// ----
	Red_Black_Green=2, // --
	Red_White_Green=3	// --
};

//------------------------------------------------------------------	class CSOM
class CSOM //  SOM
{
// 
public:
	CSOMNode m_node[]; //    
	int m_nSet; //    	
	double m_set[]; //    
	double m_max[]; //  
	double m_min[]; //  

	int	m_dimension; //   (   )
	double m_map_radius;
	double m_time_constant;
	double m_learning_rate;
	int m_iterations;
	string m_titles[];
	string m_set_titles[];

// 
public:
	uchar	Palette[768];
	int m_bmpw, m_bmph; //  BMP-
	int m_xcells, m_ycells; //  
	cIntBMP m_bmp[]; //     BMP-
	ColorSchemes m_clrSchema; //  
	int m_maxpict; //    
	bool m_bHexCell; //  
	bool m_bShowBorder; //  
	bool m_bShowTitle; //  

	int m_x0, m_y0; //   
	long m_chart; //    
	int m_wnd; //  
	string m_sID; //   .  
	
public:
	CSOM(); // 
	~CSOM(); // 
	void Init(int iter, int xc, int yc, int bmpw, int bmph, int maxpict, ColorSchemes clrSchema, bool bhex, bool bborder, bool btitle); //   

//     csv ';'.     -   
	bool LoadTrainData(string filename); //        
	bool ReadCSVData(int h); //    CSV-
	void AddVectorToTrainingSet(double &vector[], string title); //     
	void ParseString(string value,string delimiter, string  &mystrings[]); //    

// 
	void Train(bool bShowProgress); //  
	int BestMatchingNode(double &vector[]); //       

// 
	void Render(); //    
	void RenderCell(int img,int col,int node_index,bool cell_even); //   
	void ShowBMP(); //  bmp-  
	void HideBMP(); //  bmp-
	void ShowTrainPatterns(); //      
	void ShowPattern(double &vector[],string name); //       

// 
	void Blend(uchar c1,uchar c2,uchar r1,uchar g1,uchar b1,uchar r2,uchar g2,uchar b2); //    
	void ColToRGB(int col,int &r,int &g,int &b); //   col  RGB-
	int RGB256(int r,int g,int b) const {return(r+256*g+65536*b);		} //     RGB-
	int GetPalColor(uchar index); //       
	void HideChart(color BackClr); //    
};
//------------------------------------------------------------------	CSOM
void CSOM::CSOM() // 
{
	m_iterations=10; m_learning_rate=0.1; m_nSet=0;
	ArrayResize(m_set,0); ArrayResize(m_titles, 0); ArrayResize(m_set_titles, 0);
}
//------------------------------------------------------------------	~CSOM
void CSOM::~CSOM() // 
{
	ArrayResize(m_set, 0); ArrayResize(m_titles, 0); ArrayResize(m_set_titles, 0);
}
//------------------------------------------------------------------	InitParameters
void CSOM::Init(int iter, int xc, int yc, int bmpw, int bmph, int maxpict, ColorSchemes clrSchema, bool bhex=true, bool bborder=false, bool btitle=true) //   
{
	m_x0=10; m_y0=40; //   
	m_chart=ChartID(); //    
	m_wnd=0; //  
	if (m_sID=="") m_sID="SOM"; //     . 
	color clr=(color)ChartGetInteger(m_chart, CHART_COLOR_BACKGROUND);
	HideChart(clr); //  
	
	m_bHexCell=bhex; m_bShowBorder=bborder; m_bShowTitle=btitle;
	m_maxpict=maxpict; m_clrSchema=clrSchema;
	m_bmpw=bmpw; m_bmph=bmph;
	m_xcells=xc; m_ycells=yc;

	m_iterations=iter; //   	
	int cellwidth=m_bmpw/m_xcells;
	int cellheight=m_bmph/m_ycells;

	ArrayResize(m_bmp, m_dimension+1); //  BMP-
	for(int i=0; i<m_dimension; i++)
		if(!m_bHexCell) m_bmp[i].Create(m_bmpw, m_bmph, clr); else m_bmp[i].Create(m_bmpw+cellwidth, m_bmph+cellheight/2, clr);

	//
	switch(m_clrSchema)
  {
		case 0: // Black to White gradient
	  	Blend(0,255,0,0,0,255,255,255); 
	  	break;

		case 1: // Blue-green-red
			Blend(0,1*64-1,98,98,255,98,255,255);
			Blend(1*64,2*64-1,98,255,255,98,255,98);
			Blend(2*64,3*64-1,98,255,98,255,255,98);
			Blend(3*64,4*64-1,255,255,98,255,98,98);
			break;

		case 2: // Red-black-green
			Blend(0,127,155,0,0,0,0,0);
			Blend(127,255,0,0,0,0,155,0);
			break;
		case 3: // Red-white-green
			Blend(0,127,155,0,0,255,255,255);
			Blend(127,255,255,255,255,0,155,0);
			break;
		default:
			Blend(0, 255, 0, 0, 0, 255, 255, 255);
  }
  
  //     
	ArrayResize(m_node, m_xcells*m_ycells); 
	//  
	int ind=0;
	for(int i=0; i<m_xcells; i++)
		for(int j=0; j<m_ycells; j++)
	  {
			m_node[ind].SetDimension(m_dimension);
			m_node[ind].InitNode(i*cellwidth,j*cellheight,(i+1)*cellwidth,(j+1)*cellheight);
			ind++;
	  }
	//    
	if(m_bmpw>m_bmph) m_map_radius=m_bmpw/2.0; else m_map_radius=m_bmph/2.0;
	m_time_constant=1.0*m_iterations/MathLog(m_map_radius);
	Print("m_time_constant=", m_time_constant);
}

//  
//------------------------------------------------------------------	LoadTrainDataFromFile
bool CSOM::LoadTrainData(string filename) //       filename
{
	ResetLastError();
	int h=FileOpen(filename, FILE_READ|FILE_ANSI);
	if(h==INVALID_HANDLE) { Print("   ",filename," : ",GetLastError()); return(false); }
	Print("FileOpen OK");
	bool rez=ReadCSVData(h);
	FileClose(h);
	m_sID=StringSubstr(filename, 0, StringLen(filename)-4); //       
	return(rez);
}
//------------------------------------------------------------------	ReadCSVData
bool CSOM::ReadCSVData(int h)
{
	string line[];
	int n=0;
	while(!FileIsEnding(h))
  {
		string s=FileReadString(h); if(StringLen(s)<=0) continue;
		n++;
		if(n>1)
	  {
			ParseString(s, ";", line);
			double data[];
			int dim=ArraySize(line)-1; 
			if (m_dimension!=dim) { Print("-err Dimension in ",n," line"); return(false); }
			ArrayResize(data, dim);
			for(int i=0; i<dim; i++) data[i]=StringToDouble(line[i+1]);
			AddVectorToTrainingSet(data, line[0]); //     
	  }
		else
	  {
			ParseString(s, ";", line);
			m_dimension=ArraySize(line)-1;
			Print("HEADER: (",m_dimension+1,") ", s);
			//  
			ArrayResize(m_titles, m_dimension);
			for(int i=0; i<m_dimension; i++) m_titles[i]=line[i+1];
	  }
  }
  return(true);
}
//------------------------------------------------------------------	AddVectorToTrainingSet
void CSOM::AddVectorToTrainingSet(double &vector[],string title) //     
{
	m_nSet++;
	ArrayResize(m_set, m_dimension*m_nSet);
	ArrayResize(m_set_titles, m_nSet);
	m_set_titles[m_nSet-1]=title;
	for(int i=0; i<m_dimension; i++) m_set[m_dimension*(m_nSet-1)+i]=vector[i];
}
//------------------------------------------------------------------	ParseString
void CSOM::ParseString(string value, string delimiter, string &mystrings[]) //    ,  delimiter
{
	int dlen=StringLen(delimiter);
	string txt=value+delimiter;
	int strind=0;
	while(StringLen(txt)>0)
  {
		strind++;
		int indx=StringFind(txt, delimiter);
		string tempstr=StringSubstr(txt, 0, indx);
		txt=StringSubstr(txt, indx+dlen, -1);
		ArrayResize(mystrings, strind, 50);
		mystrings[strind-1]=tempstr;
  }
}

// 
//------------------------------------------------------------------	Train
void CSOM::Train(bool bShowProgress) //   
{
	double learning_rate=m_learning_rate;
	int iter=0;
	double data[];
	ArrayResize(data, m_dimension);

	Print("Dimension=", m_dimension);

	//        
	ArrayResize(m_max, m_dimension);
	ArrayResize(m_min, m_dimension);

	for(int j=0; j<m_dimension; j++)
  {
		double maxv=m_set[m_dimension+j];
		double minv=m_set[m_dimension+j];
		for(int i=1; i<m_nSet; i++)
	  {
			double v=m_set[m_dimension*i+j];
			if(v>maxv) maxv=v;
			if(v<minv) minv=v;
	  }
		m_max[j]=maxv;
		m_min[j]=minv;
		Print(j, "m_min_value=", m_min[j], "m_max_value=", m_max[j]);
  }
	int total_nodes=ArraySize(m_node);
	do
  {
		int ind=int(1.0*m_nSet*rand()/32768); //        
		//  data  
		for(int k=0; k<m_dimension; k++) data[k]=m_set[m_dimension*(ind)+k];
		int winningnode=BestMatchingNode(data); //    ,    data
		double neighbourhood_radius=m_map_radius*MathExp(-1.0*iter/m_time_constant); //    
		for(int i=0; i<total_nodes; i++) //     
	  {
			//        m_WinningNodeIndex
			double DistToNodeSqr=(m_node[winningnode].X()-m_node[i].X())*(m_node[winningnode].X()-m_node[i].X())
										+(m_node[winningnode].Y()-m_node[i].Y())*(m_node[winningnode].Y()-m_node[i].Y());
			double WS=neighbourhood_radius*neighbourhood_radius; //   
			if(DistToNodeSqr<WS) //    ,   
		  {
				double influence=MathExp(-DistToNodeSqr/(2*WS)); //     
				m_node[i].AdjustWeights(data, learning_rate, influence); //       
		  }
	  }
		learning_rate=m_learning_rate*MathExp(-1.0*iter/m_iterations); //  ( iter)   
		iter++; //   
		
		if(GetTickCount()%1000==0) //          N- 
	  {
			Comment("\n     ",iter,"  ",m_iterations," ",int(iter*100.0/m_iterations),"%");
			if (bShowProgress) { Render(); ShowBMP(); } //     
	  }
  }
	while(iter<m_iterations); //     ,       
	Comment(" ",iter,"  ",m_iterations," ",int(iter*100.0/m_iterations),"%");
	Print(" .");
}
//------------------------------------------------------------------	BestMatchingNode
int CSOM::BestMatchingNode(double &vector[]) //     
{
	int	 min_ind=0;
	double min_dist=m_node[min_ind].CalculateDistance(vector);
	int total_nodes= ArraySize(m_node);
	for(int i=0; i<total_nodes; i++)
		if(m_node[i].CalculateDistance(vector)<min_dist) 
		{ min_dist=m_node[i].CalculateDistance(vector); min_ind=i; }
	return(min_ind);
}

// 
//------------------------------------------------------------------	Render
void CSOM::Render() //      ( m_bmp)
{
	int ind=0;
	for(int i=0; i<m_xcells; i++)
  {
		for(int j=0; j<m_ycells; j++)
	  {
			//        
			for(int k=0; k<m_dimension; k++)
		  {
				double range=(m_max[k]-m_min[k]);
				char avg=char(255*(m_node[ind].GetWeight(k)-m_min[k])/range);
				int col=int(GetPalColor(avg));
				RenderCell(k,col,ind,(j%2==0));
		  }
			ind++;
	  }
		//        
		for(int m=0; m<m_dimension; m++)
	  {
			for(int k=0; k<15; k++)
				for(int l=0; l<m_bmpw; l++)
			  {
					uchar colind=uchar(MathRound(l*255.0/m_bmpw));
					int col=int(GetPalColor(colind));
					m_bmp[m].DrawDot(l,k,col);
			  }
			//         
			m_bmp[m].TypeText(0,0,DoubleToString(m_min[m],2),clrBlack);
			m_bmp[m].TypeText(m_bmpw-25,0,DoubleToString(m_max[m],2),clrBlack);

			if(m_clrSchema==2) m_bmp[m].TypeText(m_bmpw/2-3*StringLen(m_titles[m]), 0, m_titles[m], clrWhite);
			else m_bmp[m].TypeText(m_bmpw/2-3*StringLen(m_titles[m]), 0, m_titles[m], clrBlack);
	  }
  }
	return;
}
//------------------------------------------------------------------	RenderCell
void CSOM::RenderCell(int img,int col,int node_index,bool cell_even)
{
	int x1,y1,x2,y2;
	m_node[node_index].GetCoordinates(x1,y1,x2,y2);

	if(!m_bHexCell) //  
  {
		m_bmp[img].DrawRectangle(x1,y1,x2,y2,col,true);
		if(m_bShowBorder) //  
			m_bmp[img].DrawRectangle(x1,y1,x2,y2,clrBlack,false);
  }
	else // 
	{
		int x_size=int (MathAbs(x2-x1));
		int y_size=int (MathAbs(y2-y1));
		if(cell_even) { x1=x1+x_size/2; x2=x2+x_size/2; }
		y1=15+y1+y_size/4;
		y2=15+y2+y_size/4;

		m_bmp[img].DrawRectangle(x1,y1+y_size/4,x2,y2-y_size/4,col,true);
		m_bmp[img].DrawTriangle(x1,y1+y_size/4,x1+x_size/2,y1-y_size/4,x2,y1+y_size/4,col,true);
		m_bmp[img].DrawTriangle(x1,y2-y_size/4,x2,y2-y_size/4,x1+x_size/2,y2+y_size/4,col,true);

		if(m_bShowBorder) //  
	  {
			int bcol=clrBlack;
			m_bmp[img].DrawLine(x1,y1+y_size/4,x1+x_size/2,y1-y_size/4,bcol);
			m_bmp[img].DrawLine(x1+x_size/2,y1-y_size/4,x2,y1+y_size/4,bcol);

			m_bmp[img].DrawLine(x1,y2-y_size/4,x1+x_size/2,y2+y_size/4,bcol);
			m_bmp[img].DrawLine(x2,y2-y_size/4,x2-x_size/2,y2+y_size/4,bcol);

			m_bmp[img].DrawLine(x1,y1+y_size/4,x1,y2-y_size/4,bcol);
			m_bmp[img].DrawLine(x2,y1+y_size/4,x2,y2-y_size/4,bcol);
	  }
  }
}
//------------------------------------------------------------------	ShowBMP
void CSOM::ShowBMP() //   bmp-    
{
	for(int i=0; i<m_dimension; i++)
  {
  	string name=string(i)+"_"+m_titles[i];
		m_bmp[i].Save(m_sID+"\\"+name, false); //  BMP-  
		int pos1=i/m_maxpict; //    
		int pos2=i%m_maxpict;
		m_bmp[i].Show(m_x0+pos2*(15+m_bmpw), m_y0+pos1*(15+m_bmph), m_sID+"\\"+name, m_sID+name, false);
		ObjectSetInteger(0, m_sID+name, OBJPROP_BACK, false); //    
	}
}
//------------------------------------------------------------------	HideBMP
void CSOM::HideBMP()
{
	string names[]; int i, n=ObjectsTotal(m_chart, m_wnd, OBJ_BITMAP); ArrayResize(names, n);
	for (i=0; i<n; i++) names[i]=ObjectName(m_chart, i);
	for (i=0; i<n; i++) 
	{
		if (ObjectFind(m_chart, names[i])!=m_wnd) continue;
		if (ObjectGetInteger(m_chart, names[i], OBJPROP_TYPE)!=OBJ_BITMAP) continue;
		if (StringSubstr(names[i], 0, 3)!=m_sID) continue;
		ObjectDelete(m_chart, names[i]);
	}
}
//------------------------------------------------------------------	ShowPattern
void CSOM::ShowPattern(double &vector[],string name) //       					 |
{
	int WinningNodeIndex=BestMatchingNode(vector);
	int ind=BestMatchingNode(vector);
	int x1,y1,x2,y2;
	m_node[ind].GetCoordinates(x1,y1,x2,y2);
	int avgcolor=0;
	for(int i=0; i<m_dimension; i++) avgcolor+=int((255*(m_node[ind].GetWeight(i)-m_min[i])/(m_max[i]-m_min[i])));
	if(m_bHexCell) //  
  {
		int x_size=int(MathAbs(x2-x1));
		int y_size=int(MathAbs(y2-y1));
		y1=y1+y_size/4; 
		y2=y2+y_size/4;
  }
	int xc=(x1+x2)/2;
	int yc=(y1+y2)/2;
	if(m_bShowTitle) // 
	{
		for(int i=0; i<m_dimension; i++)
			if(avgcolor<=m_dimension*128)
		  {
				m_bmp[i].TypeText(xc+1,yc+1,name, clrWhite);
				m_bmp[i].TypeText(xc,yc,name, clrBlack);
		  }
			else
		  {
				m_bmp[i].TypeText(xc+1,yc+1,name, clrWhite);
				m_bmp[i].TypeText(xc,yc,name, clrBlack);
		  }
  }
//	Print(" ",name,":     =",ind);
}
//------------------------------------------------------------------	ShowTrainPatterns
void CSOM::ShowTrainPatterns() //         -
{
	double data[]; ArrayResize(data, m_dimension);
	for(int ind=0; ind<m_nSet; ind++) //  data     
  {
		for(int k=0; k<m_dimension; k++) data[k]=m_set[m_dimension*(ind)+k];
		ShowPattern(data, m_set_titles[ind]);
  }
}

//  
//------------------------------------------------------------------	Blend
void CSOM::Blend(uchar c1,uchar c2,uchar r1,uchar g1,uchar b1,uchar r2,uchar g2,uchar b2)
{
	int n=(c2-c1);
	for(int i=0; i<=n; i++)
  {
		if((c1+i+2)<ArraySize(Palette))
	  {
			Palette[3*(c1+i)]=uchar(MathRound(1*(r1*(n-i)+r2*i)*1.0/n));
			Palette[3*(c1+i)+1]=uchar (MathRound(1*(g1*(n-i) + g2*i) * 1.0/ n));
			Palette[3*(c1+i)+2]=uchar (MathRound(1*(b1*(n-i) + b2*i) * 1.0/ n));
	  }
  }
}
//------------------------------------------------------------------	GetPalColor
int CSOM::GetPalColor(uchar index) //       
{
	int ind=index; if (ind<=0) ind=0;
	uchar r=Palette[3*(ind)];
	uchar g=Palette[3*(ind)+1];
	uchar b=Palette[3*(ind)+2];
	return(RGB256(r,g,b));
}
//------------------------------------------------------------------	ColToRGB
void CSOM::ColToRGB(int col,int &r,int &g,int &b) //    col  RGB-
{
	r=col & 0xFF;
	g=(col>>8)& 0xFF;
	b=(col>>16)& 0xFF;
}
//------------------------------------------------------------------	HideChart
void CSOM::HideChart(color clrBack)
{	
	ChartSetInteger(m_chart, CHART_SHOW_OHLC, false);
	ChartSetInteger(m_chart, CHART_SHOW_BID_LINE, false);
	ChartSetInteger(m_chart, CHART_SHOW_ASK_LINE, false);
	ChartSetInteger(m_chart, CHART_SHOW_LAST_LINE, false);
	ChartSetInteger(m_chart, CHART_SHOW_PERIOD_SEP, false);
	ChartSetInteger(m_chart, CHART_SHOW_GRID, false);
	ChartSetInteger(m_chart, CHART_SHOW_VOLUMES, CHART_VOLUME_HIDE);
	ChartSetInteger(m_chart, CHART_SHOW_OBJECT_DESCR, false);
	ChartSetInteger(m_chart, CHART_SHOW_TRADE_LEVELS, false);
	ChartSetInteger(m_chart, CHART_SHOW_DATE_SCALE, false);
	ChartSetInteger(m_chart, CHART_SHOW_PRICE_SCALE, false);
	ChartSetInteger(m_chart, CHART_FOREGROUND, false);
	ChartSetInteger(m_chart, CHART_SHIFT, false);
	ChartSetInteger(m_chart, CHART_MODE, CHART_LINE);
	ChartSetInteger(m_chart, CHART_SCALE, 0);
	
	ChartSetInteger(m_chart, CHART_COLOR_BACKGROUND, clrBack);
	ChartSetInteger(m_chart, CHART_COLOR_FOREGROUND, clrWhite);
	ChartSetInteger(m_chart, CHART_COLOR_CHART_LINE, clrBack);
	ChartSetInteger(m_chart, CHART_COLOR_ASK, clrBack);
	ChartSetInteger(m_chart, CHART_COLOR_BID, clrBack);
	ChartSetInteger(m_chart, CHART_SCALEFIX, true);
	ChartSetDouble(m_chart, CHART_FIXED_MAX, 1);
	ChartSetDouble(m_chart, CHART_FIXED_MIN, 1);
	ChartSetString(m_chart, CHART_COMMENT, "");
	ChartRedraw(m_chart);
}